from django.utils.translation import gettext as _

EMAIL_CONFIRMATION_PIN_TITLE = _("Email confirmation PIN")
EMAIL_CONFIRMATION_PIN_BODY = _("Your email confirmation PIN is %s")
PHONE_NUMBER_CONFIRMATION_PIN_BODY = _("Your email confirmation PIN is %s")
PASSWORD_RESET_PIN_BODY = _("Your email confirmation PIN is %s")
PASSWORD_RESET_PIN_TITLE = _("Password reset")


from typing import Dict, List

from rest_framework import status
from rest_framework.exceptions import APIException


class KwargsError(object):
    code: str
    kwargs: Dict[str, str]

    def __init__(self, code: str, kwargs: Dict[str, str] | None = None):
        self.code = code
        self.kwargs = kwargs


class CustomAPIException(APIException):
    default_detail = "Invalid input."
    default_code = "error"

    def __init__(self, errors: Dict[str, List[KwargsError]], status_code=status.HTTP_400_BAD_REQUEST):
        """
        :param errors: List of error dictionaries, each with `code` and optional `kwargs`.
        """
        if not isinstance(errors, dict):
            raise ValueError("Errors must be a dict.")
        self.detail = 'Bad request'
        self.status_code = status_code
        self.errors_repr = {
            field_name: [
                {
                    'error_code': error.code,
                    'kwargs': error.kwargs,
                }
                for error in errors
            ]
            for field_name, errors in errors.items()
        }



from rest_framework.renderers import JSONRenderer


class CustomJSONRenderer(JSONRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        # If data already has the expected structure, return it as-is
        if isinstance(data, dict) and "errors" in data and "data" in data:
            return super().render(data, accepted_media_type, renderer_context)

        response_data = {"errors": {}, "data": {}}

        if renderer_context:
            response = renderer_context["response"]
            if response.status_code < 300:
                if response.status_code == 204:
                    response_data["data"] = None
                else:
                    response_data["data"] = data
            else:
                response_data["errors"] = data if data else {}
        else:
            response_data["data"] = data

        return super().render(response_data, accepted_media_type, renderer_context)



INVALID_CONFIRMATION_CODE = 'invalid_confirmation_code'
SITE_WITH_DOMAIN_DOES_NOT_EXIST = 'site_with_domain_does_not_exist'
INCORRECT_PHONE_NUMBER = 'incorrect_phone_number'
INCORRECT_CONFIRMATION_PIN = 'incorrect_confirmation_pin'
EMAIL_ALREADY_REGISTERED_ON_DOMAIN = 'email_already_registered_on_domain'
PHONE_NUMBER_ALREADY_REGISTERED_ON_DOMAIN = 'phone_number_already_registered_on_domain'
INVALID_PASSWORD = 'invalid_password'
CREDENTIAL_NOT_CONFIRMED = 'credential_not_confirmed'
USER_WITH_PROVIDED_CREDENTIALS_DOES_NOT_REGISTERED_ON_SPECIFIED_DOMAIN = 'user_with_provided_credentials_does_not_registered_on_specified_domain'
TOO_FREQUENT_SMS_REQUESTS = 'too_frequent_sms_requests'
INVALID_PHONE_NUMBER = 'invalid_phone_number'
INVALID_OTP = 'invalid_otp'
INVALID_CREDENTIAL = 'invalid_credential'
INVALID_CREDENTIAL_TYPE = 'invalid_credential_type'
ATTEMPTS_LIMIT_REACHED = 'attempts_limit_reached'
INVALID_SMS_TYPE = 'invalid_sms_type'
PHONE_NUMBER_NOT_CONFIRMED = 'phone_number_not_confirmed'

# Google Sign-In error codes
GOOGLE_SIGN_IN_NOT_CONFIGURED = 'google_sign_in_not_configured'
INVALID_GOOGLE_ID_TOKEN = 'invalid_google_id_token'
INVALID_GOOGLE_AUTH_TEMP_TOKEN = 'invalid_google_auth_temp_token'
GOOGLE_AUTH_TEMP_TOKEN_EXPIRED = 'google_auth_temp_token_expired'

# Validator error codes
INVALID_EMAIL = 'invalid_email'
FIELD_IS_REQUIRED = 'field_is_required'
PASSWORD_TOO_SHORT = 'password_too_short'
PASSWORD_TOO_COMMON = 'password_too_common'
PASSWORD_ENTIRELY_NUMERIC = 'password_entirely_numeric'
PASSWORD_TOO_SIMILAR = 'password_too_similar'



from functools import wraps

from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.response import Response

from moses.services.mfa import check_mfa_otp


def request_passes_test(test_func):
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(view, request, *args, **kwargs):
            if test_func(request):
                return view_func(view, request, *args, **kwargs)
            return Response({'error': _("OTP not provided")}, status.HTTP_401_UNAUTHORIZED)
        return _wrapped_view
    return decorator


def otp_required(function=None):
    """
    Decorator for views that checks that the user is logged in, redirecting
    to the log-in page if necessary.
    """
    actual_decorator = request_passes_test(
        lambda request: check_mfa_otp(request.user, request.headers.get('OTP')),
    )
    if function:
        return actual_decorator(function)
    return actual_decorator


from django.urls import path
from rest_framework.routers import DefaultRouter
from rest_framework_simplejwt.views import (
    TokenRefreshView
)

from .views.google_auth import GoogleSignInView, GoogleCompleteRegistrationView
from .views.token_obtain_pair import TokenObtainPairView
from .views.user import UserViewSet

router = DefaultRouter()
router.register("users", UserViewSet)

app_name = 'moses'

urlpatterns = [
                  path('token/obtain/', TokenObtainPairView.as_view(), name='token_obtain_pair'),
                  path('token/refresh/', TokenRefreshView.as_view(), name='token_refresh'),
                  path('token/google/', GoogleSignInView.as_view(), name='google_sign_in'),
                  path('token/google/complete/', GoogleCompleteRegistrationView.as_view(), name='google_complete_registration'),
              ] + router.urls




import random
from datetime import timedelta

from django.core.mail import send_mail
from django.utils import translation, timezone
from django.utils.timezone import now

from moses.common import error_codes
from moses.common.exceptions import CustomAPIException, KwargsError
from moses.conf import settings as moses_settings
from moses.constants import strings
from moses.enums import Credential, SMSType
from moses.services.sms import sms_unlock_time
from moses.signals import phone_number_confirmed, email_confirmed


def send_email_confirmation_message(email: str, body: str):
    send_mail(str(strings.EMAIL_CONFIRMATION_PIN_TITLE), body, moses_settings.SENDER_EMAIL, [email])


def try_to_confirm_credential(user, credential: Credential, main_pin_str: str, candidate_pin_str: str):
    match credential:
        case Credential.PHONE_NUMBER:
            attempts_field = 'phone_number_confirmation_attempts'
            current_credential_field = 'phone_number'
            candidate_credential_field = 'phone_number_candidate'
            current_credential_confirmation_field = 'is_phone_number_confirmed'
            max_attempts_limit = moses_settings.PHONE_NUMBER_CONFIRMATION_ATTEMPTS_LIMIT
            current_pin_field = 'phone_number_confirmation_pin'
            candidate_pin_field = 'phone_number_candidate_confirmation_pin'
        case Credential.EMAIL:
            attempts_field = 'email_confirmation_attempts'
            current_credential_field = 'email'
            candidate_credential_field = 'email_candidate'
            current_credential_confirmation_field = 'is_email_confirmed'
            max_attempts_limit = moses_settings.EMAIL_CONFIRMATION_ATTEMPTS_LIMIT
            current_pin_field = 'email_confirmation_pin'
            candidate_pin_field = 'email_candidate_confirmation_pin'
        case _:
            raise ValueError(error_codes.INVALID_CREDENTIAL)
    if getattr(user, attempts_field) >= max_attempts_limit:
        raise CustomAPIException(
            {
                '': [
                    KwargsError(
                        code=error_codes.ATTEMPTS_LIMIT_REACHED)
                ]
            }
        )
    received_pin, received_candidate_pin = int(main_pin_str or '0'), int(candidate_pin_str or '0')
    is_main_pin_correct = received_pin == getattr(user, current_pin_field)
    is_candidate_pin_correct = None
    if getattr(user, candidate_credential_field):
        is_candidate_pin_correct = getattr(user, candidate_pin_field) == received_candidate_pin
    if is_main_pin_correct and (is_candidate_pin_correct is None or is_candidate_pin_correct):
        setattr(user, current_pin_field, 0)
        setattr(user, candidate_pin_field, 0)
        setattr(user, current_credential_confirmation_field, True)
        setattr(user, attempts_field, 0)

        # Check if this is an update (had candidate) or initial confirmation
        had_candidate = bool(getattr(user, candidate_credential_field))

        if candidate := getattr(user, candidate_credential_field):
            setattr(user, current_credential_field, candidate)
            setattr(user, candidate_credential_field, '')
        user.save()

        # Emit signal after successful confirmation
        confirmed_value = getattr(user, current_credential_field)
        is_initial_confirmation = not had_candidate

        if credential == Credential.PHONE_NUMBER:
            phone_number_confirmed.send(
                sender=user.__class__,
                user=user,
                phone_number=confirmed_value,
                is_initial_confirmation=is_initial_confirmation
            )
        elif credential == Credential.EMAIL:
            email_confirmed.send(
                sender=user.__class__,
                user=user,
                email=confirmed_value,
                is_initial_confirmation=is_initial_confirmation
            )
    else:
        setattr(user, attempts_field, getattr(user, attempts_field) + 1)
        user.save()
    return is_main_pin_correct, is_candidate_pin_correct


def send_credential_confirmation_code(
        user,
        credential_type:
        Credential,
        candidate=False,
        generate_new=False,
        ignore_frequency_limit=False
):
    match credential_type:
        case Credential.PHONE_NUMBER:
            send_function = moses_settings.SEND_SMS_HANDLER
            message_body = str(strings.PHONE_NUMBER_CONFIRMATION_PIN_BODY)
            if candidate:
                unlock_time_field = 'phone_number_candidate_confirmation_code_sms_unlocks_at'
                credential_field = 'phone_number_candidate'
                pin_field = 'phone_number_candidate_confirmation_pin'
            else:
                unlock_time_field = 'phone_number_confirmation_code_sms_unlocks_at'
                credential_field = 'phone_number'
                pin_field = 'phone_number_confirmation_pin'
            if (
                    (
                            sut := sms_unlock_time(
                                user,
                                SMSType.PHONE_NUMBER_CONFIRMATION,
                                candidate=candidate)
                    ) is None or sut <= timezone.now() or ignore_frequency_limit
            ):
                new_unlock_time = now() + timedelta(minutes=moses_settings.PHONE_NUMBER_CONFIRMATION_SMS_MINUTES_PERIOD)
                setattr(user, unlock_time_field, new_unlock_time)
            else:
                raise CustomAPIException(
                    {
                        '': [
                            KwargsError(code=error_codes.TOO_FREQUENT_SMS_REQUESTS)
                        ]
                    }
                )
        case Credential.EMAIL:
            send_function = send_email_confirmation_message
            message_body = str(strings.EMAIL_CONFIRMATION_PIN_BODY)
            if candidate:
                credential_field = 'email_candidate'
                pin_field = 'email_candidate_confirmation_pin'
            else:
                credential_field = 'email'
                pin_field = 'email_confirmation_pin'
        case _:
            raise ValueError(error_codes.INVALID_CREDENTIAL_TYPE)

    if generate_new:
        setattr(user, pin_field, random.randint(100000, 999999))
    user.save()
    with translation.override(user.preferred_language):
        send_function(getattr(user, credential_field), message_body % getattr(user, pin_field))
